# cython: language_level=3

# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from numpy import asarray, ascontiguousarray, empty, linspace
from cherab.core.math.function cimport Function1D, Function2D, Function3D, VectorFunction2D, VectorFunction3D
from cherab.core.math.function cimport autowrap_function1d, autowrap_function2d, autowrap_function3d, autowrap_vectorfunction2d, autowrap_vectorfunction3d
from raysect.core cimport Vector3D
cimport cython
cimport numpy as np

"""
This module provides a set of sampling functions for rapidly generating samples
of 1D, 2D and 3D functions.

These functions use C calls when sampling Function1D, Function2D and Function3D
objects and are therefore considerably faster than the equivalent Python code.
"""

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple sample1d(object function1d, tuple x_range):
    """
    Samples a 1D function over the specified range.

    :param function1d: a Python function or Function1D object
    :param x_range: a tuple defining the sample range: (min, max, samples)
    :return: a tuple containing the sampled values: (x_points, function_samples)
    
    .. code-block:: pycon

       >>> from cherab.core.math import sample1d
       >>>
       >>> def f1(x):
       >>>     return x**2
       >>>
       >>> x_pts, f_vals = sample1d(f1, (0, 3, 5))
       >>> x_pts
       array([0.  , 0.75, 1.5 , 2.25, 3.  ])
       >>> f_vals
       array([0.    , 0.5625, 2.25  , 5.0625, 9.    ])
    """

    cdef:
        int i, samples
        Function1D f1d
        double[::1] x_view, v_view

    if len(x_range) != 3:
        raise ValueError("Range must be a tuple containing: (min range, max range, no. of samples).")

    if x_range[0] > x_range[1]:
        raise ValueError("Minimum range can not be greater than maximum range.")

    if x_range[2] < 1:
        raise ValueError("The number of samples must be >= 1.")

    f1d = autowrap_function1d(function1d)
    samples = x_range[2]

    x = linspace(x_range[0], x_range[1], samples)
    v = empty(samples)

    # obtain memoryviews for fast, direct memory access
    x_view = x
    v_view = v

    for i in range(samples):
        v_view[i] = f1d.evaluate(x_view[i])

    return x, v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray sample1d_points(object function1d, object x_points):
    """
    Sample a 1D function at the specified points

    :param function1d: a Python function or Function1D object
    :param x_points: an array of points at which to sample the function
    :return: an array containing the sampled values

    .. code-block:: pycon

       >>> from cherab.core.math import sample1d_points
       >>>
       >>> def f1(x):
       >>>     return x**2
       >>>
       >>> f_vals = sample1d_points(f1, [1, 2, 3])
       >>> f_vals
       array([1., 4., 9.])
    """
    cdef:
        int i, nsamples
        Function1D f1d
        double[::1] x_view, v_view

    x_points = ascontiguousarray(x_points, dtype=float)

    f1d = autowrap_function1d(function1d)
    nsamples = len(x_points)

    v = empty(nsamples)

    x_view = x_points
    v_view = v

    for i in range(nsamples):
        v_view[i] = f1d.evaluate(x_view[i])

    return v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple sample2d(object function2d, tuple x_range, tuple y_range):
    """
    Samples a 2D function over the specified range.

    :param function2d: a Python function or Function2D object
    :param x_range: a tuple defining the x sample range: (x_min, x_max, x_samples)
    :param y_range: a tuple defining the y sample range: (y_min, y_max, y_samples)
    :return: a tuple containing the sampled values: (x_points, y_points, function_samples)

    .. code-block:: pycon

       >>> from cherab.core.math import sample2d
       >>>
       >>> def f1(x, y):
       >>>     return x**2 + y
       >>>
       >>> x_pts, y_pts, f_vals = sample2d(f1, (1, 3, 5), (1, 3, 5))
       >>> x_pts
       array([1. , 1.5, 2. , 2.5, 3. ])
       >>> y_pts
       array([1. , 1.5, 2. , 2.5, 3. ])
       >>> f_vals
       array([[ 2.  ,  2.5 ,  3.  ,  3.5 ,  4.  ],
              [ 3.25,  3.75,  4.25,  4.75,  5.25],
              [ 5.  ,  5.5 ,  6.  ,  6.5 ,  7.  ],
              [ 7.25,  7.75,  8.25,  8.75,  9.25],
              [10.  , 10.5 , 11.  , 11.5 , 12.  ]])
    """

    cdef:
        int i, j
        Function2D f2d
        int x_samples, y_samples
        double[::1] x_view, y_view
        double[:, ::1] v_view

    if len(x_range) != 3:
        raise ValueError("X range must be a tuple containing: (min range, max range, no. of samples).")

    if len(y_range) != 3:
        raise ValueError("Y range must be a tuple containing: (min range, max range, no. of samples).")

    if x_range[0] > x_range[1]:
        raise ValueError("Minimum x range can not be greater than maximum x range.")

    if y_range[0] > y_range[1]:
        raise ValueError("Minimum y range can not be greater than maximum y range.")

    if x_range[2] < 1:
        raise ValueError("The number of x samples must be >= 1.")

    if y_range[2] < 1:
        raise ValueError("The number of y samples must be >= 1.")

    f2d = autowrap_function2d(function2d)
    x_samples = x_range[2]
    y_samples = y_range[2]

    x = linspace(x_range[0], x_range[1], x_samples)
    y = linspace(y_range[0], y_range[1], y_samples)
    v = empty((x_samples, y_samples))

    # obtain memoryviews for fast, direct memory access
    x_view = x
    y_view = y
    v_view = v

    for i in range(x_samples):
        for j in range(y_samples):
            v_view[i, j] = f2d.evaluate(x_view[i], y_view[j])

    return x, y, v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray sample2d_points(object function2d, object points):
    """
    Sample a 2D function at the specified points.

    This function is for sampling at an unstructured sequence of points.
    For sampling over a regular grid, consider sample2d or sample2d_grid
    instead.

    :param function2d: a Python function or Function2D object
    :param points: an Nx2 array of points at which to sample the function
    :return: a 1D array containing the sampled values at each point

    .. code-block:: pycon

       >>> from cherab.core.math import sample2d
       >>>
       >>> def f1(x, y):
       >>>     return x**2 + y
       >>>
       >>> f_vals = sample2d_points(f1, [[1, 1], [2, 2], [3, 3]])
       >>> f_vals
       array([ 2.,  6., 12.])
    """
    cdef:
        int i, j, nsamples
        Function2D f2d
        double[::1] x_view, y_view, v_view

    points = asarray(points)
    if points.ndim != 2 or points.shape[1] != 2:
        raise ValueError("points should be an Nx2 array of points.")

    f2d = autowrap_function2d(function2d)
    x = ascontiguousarray(points[:, 0], dtype=float)
    y = ascontiguousarray(points[:, 1], dtype=float)
    nsamples = points.shape[0]
    v = empty(nsamples)

    x_view = x
    y_view = y
    v_view = v

    for i in range(nsamples):
        v_view[i] = f2d.evaluate(x_view[i], y_view[i])

    return v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray sample2d_grid(object function2d, object x, object y):
    """
    Sample a 2D function on a rectilinear grid

    Note that v[i, j] = f(x[i], y[j])

    :param function2d: a Python function or Function2D object
    :param x: the x coordinates of each column in the grid
    :param y: the y coordinates of each row in the grid
    :return v: a 2D array containing the sampled values at each grid point

    .. code-block:: pycon

       >>> from cherab.core.math import sample2d_grid
       >>>
       >>> def f1(x, y):
       >>>     return x**2 + y
       >>>
       >>> f_vals = sample2d_grid(f1, [1, 2], [1, 2, 3])
       >>> f_vals
       array([[2., 3., 4.],
              [5., 6., 7.]])
    """
    cdef:
        int i, j, x_samples, y_samples
        Function2D f2d
        double[::1] x_view, y_view
        double[:, ::1] v_view

    x = ascontiguousarray(x, dtype=float)
    y = ascontiguousarray(y, dtype=float)
    if x.ndim != 1:
        raise ValueError("x should be a 1D sequence of coordinates")
    if y.ndim != 1:
        raise ValueError("y should be a 1D sequence of coordinates")

    f2d = autowrap_function2d(function2d)

    x_samples = x.shape[0]
    y_samples = y.shape[0]
    v = empty((x_samples, y_samples))

    x_view = x
    y_view = y
    v_view = v

    for i in range(x_samples):
        for j in range(y_samples):
            v_view[i, j] = f2d.evaluate(x_view[i], y_view[j])

    return v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple sample3d(object function3d, tuple x_range, tuple y_range, tuple z_range):
    """
    Samples a 3D function over the specified range.

    :param function3d: a Python function or Function2D object
    :param x_range: a tuple defining the x sample range: (x_min, x_max, x_samples)
    :param y_range: a tuple defining the y sample range: (y_min, y_max, y_samples)
    :param z_range: a tuple defining the z sample range: (z_min, z_max, z_samples)
    :return: a tuple containing the sampled values: (x_points, y_points, z_points, function_samples)

    .. code-block:: pycon

       >>> from cherab.core.math import sample3d
       >>>
       >>> def f1(x, y, z):
       >>>     return x**3 + y**2 + z
       >>>
       >>> x_pts, y_pts, z_pts, f_vals = sample3d(f1, (1, 3, 3), (1, 3, 3), (1, 3, 3))
       >>> x_pts
       array([1., 2., 3.])
       >>> f_vals
       array([[[ 3.,  4.,  5.],
               [ 6.,  7.,  8.],
               [11., 12., 13.]],       
              [[10., 11., 12.],
               [13., 14., 15.],
               [18., 19., 20.]],
              [[29., 30., 31.],
               [32., 33., 34.],
               [37., 38., 39.]]])
    """

    cdef:
        int i, j, k
        Function3D f3d
        int x_samples, y_samples, z_samples
        double[::1] x_view, y_view, z_view
        double[:, :, ::1] v_view

    if len(x_range) != 3:
        raise ValueError("X range must be a tuple containing: (min range, max range, no. of samples).")

    if len(y_range) != 3:
        raise ValueError("Y range must be a tuple containing: (min range, max range, no. of samples).")

    if len(z_range) != 3:
        raise ValueError("Z range must be a tuple containing: (min range, max range, no. of samples).")

    if x_range[0] > x_range[1]:
        raise ValueError("Minimum x range can not be greater than maximum x range.")

    if y_range[0] > y_range[1]:
        raise ValueError("Minimum y range can not be greater than maximum y range.")

    if z_range[0] > z_range[1]:
        raise ValueError("Minimum z range can not be greater than maximum z range.")

    if x_range[2] < 1:
        raise ValueError("The number of x samples must be >= 1.")

    if y_range[2] < 1:
        raise ValueError("The number of y samples must be >= 1.")

    if z_range[2] < 1:
        raise ValueError("The number of z samples must be >= 1.")

    f3d = autowrap_function3d(function3d)
    x_samples = x_range[2]
    y_samples = y_range[2]
    z_samples = z_range[2]

    x = linspace(x_range[0], x_range[1], x_samples)
    y = linspace(y_range[0], y_range[1], y_samples)
    z = linspace(z_range[0], z_range[1], z_samples)
    v = empty((x_samples, y_samples, z_samples))

    # obtain memoryviews for fast, direct memory access
    x_view = x
    y_view = y
    z_view = z
    v_view = v

    for i in range(x_samples):
        for j in range(y_samples):
            for k in range(z_samples):
                v_view[i, j, k] = f3d.evaluate(x_view[i], y_view[j], z_view[k])

    return x, y, z, v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray sample3d_points(object function3d, object points):
    """
    Sample a 3D function at the specified points.

    This function is for sampling at an unstructured sequence of points.
    For sampling over a regular grid, consider sample3d or sample3d_grid
    instead.

    :param function3d: a Python function or Function3D object
    :param points: an Nx3 array of points at which to sample the function
    :return: a 1D array containing the sampled values at each point
    
    .. code-block:: pycon

       >>> from cherab.core.math import sample3d_points
       >>>
       >>> def f1(x, y, z):
       >>>     return x**3 + y**2 + z
       >>>
       >>> f_vals = sample3d_points(f1, [[1,1,1], [2,2,2], [3,3,3]])
       >>> f_vals
       array([ 3., 14., 39.])
    """
    cdef:
        int i, j, nsamples
        Function3D f3d
        double[::1] x_view, y_view, z_view, v_view

    points = asarray(points)
    if points.ndim != 2 or points.shape[1] != 3:
        raise ValueError("points should be an Nx3 array of points.")

    f3d = autowrap_function3d(function3d)
    x = ascontiguousarray(points[:, 0], dtype=float)
    y = ascontiguousarray(points[:, 1], dtype=float)
    z = ascontiguousarray(points[:, 2], dtype=float)
    nsamples = points.shape[0]
    v = empty(nsamples)

    x_view = x
    y_view = y
    z_view = z
    v_view = v

    for i in range(nsamples):
        v_view[i] = f3d.evaluate(x_view[i], y_view[i], z_view[i])

    return v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray sample3d_grid(object function3d, object x, object y, object z):
    """
    Sample a 3D function on a rectilinear grid

    Note that v[i, j, k] = f(x[i], y[j], z[k])

    :param function3d: a Python function or Function3D object
    :param x: the x coordinates of each column in the grid
    :param y: the y coordinates of each row in the grid
    :param z: the z coordinates of each plane in the grid
    :return v: a 3D array containing the sampled values at each grid point

    .. code-block:: pycon

       >>> from cherab.core.math import sample3d_grid
       >>>
       >>> def f1(x, y, z):
       >>>     return x**3 + y**2 + z
       >>>
       >>> f_vals = sample3d_grid(f1, (1, 3, 3), (1, 3, 3), (1, 3, 3))
       >>> f_vals
       array([[[ 3.,  5.,  5.],
               [11., 13., 13.],
               [11., 13., 13.]],
              [[29., 31., 31.],
               [37., 39., 39.],
               [37., 39., 39.]],
              [[29., 31., 31.],
               [37., 39., 39.],
               [37., 39., 39.]]])
    """
    cdef:
        int i, j, k, x_samples, y_samples, z_samples
        Function3D f3d
        double[::1] x_view, y_view, z_view
        double[:, :, ::1] v_view

    x = ascontiguousarray(x, dtype=float)
    y = ascontiguousarray(y, dtype=float)
    z = ascontiguousarray(z, dtype=float)
    if x.ndim != 1:
        raise ValueError("x should be a 1D sequence of coordinates")
    if y.ndim != 1:
        raise ValueError("y should be a 1D sequence of coordinates")
    if z.ndim != 1:
        raise ValueError("z should be a 1D sequence of coordinates")

    f3d = autowrap_function3d(function3d)

    x_samples = x.shape[0]
    y_samples = y.shape[0]
    z_samples = z.shape[0]
    v = empty((x_samples, y_samples, z_samples))

    x_view = x
    y_view = y
    z_view = z
    v_view = v

    for i in range(x_samples):
        for j in range(y_samples):
            for k in range(z_samples):
                v_view[i, j, k] = f3d.evaluate(x_view[i], y_view[j], z_view[k])

    return v


# todo: add test
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple samplevector2d(object function2d, tuple x_range, tuple y_range):
    """
    Samples a 2D vector function over the specified range.

    The function samples returns are an NxMx3 array where the last axis are the
    x, y, and z components of the vector respectively.

    :param function2d: a Python function or Function2D object
    :param x_range: a tuple defining the x sample range: (x_min, x_max, x_samples)
    :param y_range: a tuple defining the y sample range: (y_min, y_max, y_samples)
    :return: a tuple containing the sampled values: (x_points, y_points, function_samples)

    .. code-block:: pycon

       >>> from raysect.core.math import Vector3D
       >>> from cherab.core.math import samplevector2d
       >>>
       >>> def my_func(x, y):
       >>>     return Vector3D(x, y, 0)
       >>>
       >>> x_pts, y_pts, f_vals = samplevector2d(my_func, (1, 3, 3), (1, 3, 3))
       >>> x_pts
       array([1., 2., 3.])
       >>> y_pts
       array([1., 2., 3.])
       >>> f_vals
       array([[[1., 1., 0.],
               [1., 2., 0.],
               [1., 3., 0.]],
              [[2., 1., 0.],
               [2., 2., 0.],
               [2., 3., 0.]],
              [[3., 1., 0.],
               [3., 2., 0.],
               [3., 3., 0.]]])
    """

    cdef:
        int i, j
        VectorFunction2D f2d
        int x_samples, y_samples
        double[::1] x_view, y_view
        double[:, :, ::1] v_view
        Vector3D vector

    if len(x_range) != 3:
        raise ValueError("X range must be a tuple containing: (min range, max range, no. of samples).")

    if len(y_range) != 3:
        raise ValueError("Y range must be a tuple containing: (min range, max range, no. of samples).")

    if x_range[0] > x_range[1]:
        raise ValueError("Minimum x range can not be greater than maximum x range.")

    if y_range[0] > y_range[1]:
        raise ValueError("Minimum y range can not be greater than maximum y range.")

    if x_range[2] < 1:
        raise ValueError("The number of x samples must be >= 1.")

    if y_range[2] < 1:
        raise ValueError("The number of y samples must be >= 1.")

    f2d = autowrap_vectorfunction2d(function2d)
    x_samples = x_range[2]
    y_samples = y_range[2]

    x = linspace(x_range[0], x_range[1], x_samples)
    y = linspace(y_range[0], y_range[1], y_samples)
    v = empty((x_samples, y_samples, 3))

    # obtain memoryviews for fast, direct memory access
    x_view = x
    y_view = y
    v_view = v

    for i in range(x_samples):
        for j in range(y_samples):
            vector = f2d.evaluate(x_view[i], y_view[j])
            v_view[i, j, 0] = vector.x
            v_view[i, j, 1] = vector.y
            v_view[i, j, 2] = vector.z

    return x, y, v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray samplevector2d_points(object function2d, object points):
    """
    Sample a 2D vector function at the specified points.

    This function is for sampling at an unstructured sequence of points.
    For sampling over a regular grid, consider samplevector2d or
    samplevector2d_grid instead.

    :param function2d: a Python function or Function2D object
    :param points: an Nx2 array of points at which to sample the function
    :return: a Nx3 array containing the sampled values at each point

    .. code-block:: pycon

       >>> from raysect.core.math import Vector3D
       >>> from cherab.core.math import samplevector2d_points
       >>>
       >>> def my_func(x, y):
       >>>     return Vector3D(x, y, 0)
       >>>
       >>> f_vals = samplevector2d_points(my_func, [[1, 1], [2, 2], [3, 3]])
       >>> f_vals
       array([[1., 1., 0.],
              [2., 2., 0.],
              [3., 3., 0.]])
    """
    cdef:
        int i, j, nsamples
        VectorFunction2D f2d
        double[::1] x_view, y_view,
        double [:, ::1] v_view
        Vector3D vector

    points = asarray(points)
    if points.ndim != 2 or points.shape[1] != 2:
        raise ValueError("points should be an Nx2 array of points.")

    f2d = autowrap_vectorfunction2d(function2d)
    x = ascontiguousarray(points[:, 0], dtype=float)
    y = ascontiguousarray(points[:, 1], dtype=float)
    nsamples = points.shape[0]
    v = empty((nsamples, 3))

    x_view = x
    y_view = y
    v_view = v

    for i in range(nsamples):
        vector = f2d.evaluate(x_view[i], y_view[i])
        v_view[i, 0] = vector.x
        v_view[i, 1] = vector.y
        v_view[i, 2] = vector.z

    return v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray samplevector2d_grid(object function2d, object x, object y):
    """
    Sample a 2D vector function on a rectilinear grid

    :param function2d: a Python function or Function2D object
    :param x: the x coordinates of each column in the grid
    :param y: the y coordinates of each row in the grid
    :return v: a 3D array containing the sampled values at each grid point

    Note that v[i, j] = f(x[i], y[j])

    .. code-block:: pycon

       >>> from raysect.core.math import Vector3D
       >>> from cherab.core.math import samplevector2d_grid
       >>>
       >>> def my_func(x, y):
       >>>     return Vector3D(x, y, 0)
       >>>
       >>> f_vals = samplevector2d_grid(my_func, [1, 2], [1, 2, 3])
       >>> f_vals
       array([[[1., 1., 0.],
               [1., 2., 0.],
               [1., 3., 0.]],
              [[2., 1., 0.],
               [2., 2., 0.],
               [2., 3., 0.]]])
    """
    cdef:
        int i, j, x_samples, y_samples
        VectorFunction2D f2d
        double[::1] x_view, y_view
        double[:, :, ::1] v_view
        Vector3D vector

    x = ascontiguousarray(x, dtype=float)
    y = ascontiguousarray(y, dtype=float)
    if x.ndim != 1:
        raise ValueError("x should be a 1D sequence of coordinates")
    if y.ndim != 1:
        raise ValueError("y should be a 1D sequence of coordinates")

    f2d = autowrap_vectorfunction2d(function2d)

    x_samples = x.shape[0]
    y_samples = y.shape[0]
    v = empty((x_samples, y_samples, 3))

    x_view = x
    y_view = y
    v_view = v

    for i in range(x_samples):
        for j in range(y_samples):
            vector = f2d.evaluate(x_view[i], y_view[j])
            v_view[i, j, 0] = vector.x
            v_view[i, j, 1] = vector.y
            v_view[i, j, 2] = vector.z

    return v


# todo: add test
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple samplevector3d(object function3d, tuple x_range, tuple y_range, tuple z_range):
    """
    Samples a 3D vector function over the specified range.

    The function samples returns are an NxMxKx3 array where the last axis are the
    x, y, and z components of the vector respectively.

    :param function3d: a Python function or Function2D object
    :param x_range: a tuple defining the x sample range: (x_min, x_max, x_samples)
    :param y_range: a tuple defining the y sample range: (y_min, y_max, y_samples)
    :param z_range: a tuple defining the z sample range: (z_min, z_max, z_samples)
    :return: a tuple containing the sampled values: (x_points, y_points, z_points, function_samples)

    .. code-block:: pycon

       >>> from raysect.core.math import Vector3D
       >>> from cherab.core.math import samplevector3d
       >>>
       >>> def my_func(x, y, z):
       >>>     return Vector3D(x, y, z)
       >>>
       >>> x_pts, y_pts, z_pts, f_vals = samplevector3d(my_func, (1, 2, 2), (1, 3, 3), (1, 3, 3))
       >>> x_pts
       array([1., 2.])
       >>> f_vals
       array([[[[1., 1., 1.],
                [1., 1., 2.],
                [1., 1., 3.]],
               [[1., 2., 1.],
                [1., 2., 2.],
                [1., 2., 3.]],
               [[1., 3., 1.],
                [1., 3., 2.],
                [1., 3., 3.]]],

              [[[2., 1., 1.],
                [2., 1., 2.],
                [2., 1., 3.]],
               [[2., 2., 1.],
                [2., 2., 2.],
                [2., 2., 3.]],
               [[2., 3., 1.],
                [2., 3., 2.],
                [2., 3., 3.]]]])
    """

    cdef:
        int i, j, k
        VectorFunction3D f3d
        int x_samples, y_samples, z_samples
        double[::1] x_view, y_view, z_view
        double[:, :, :, ::1] v_view
        Vector3D vector

    if len(x_range) != 3:
        raise ValueError("X range must be a tuple containing: (min range, max range, no. of samples).")

    if len(y_range) != 3:
        raise ValueError("Y range must be a tuple containing: (min range, max range, no. of samples).")

    if len(z_range) != 3:
        raise ValueError("Z range must be a tuple containing: (min range, max range, no. of samples).")

    if x_range[0] > x_range[1]:
        raise ValueError("Minimum x range can not be greater than maximum x range.")

    if y_range[0] > y_range[1]:
        raise ValueError("Minimum y range can not be greater than maximum y range.")

    if z_range[0] > z_range[1]:
        raise ValueError("Minimum z range can not be greater than maximum z range.")

    if x_range[2] < 1:
        raise ValueError("The number of x samples must be >= 1.")

    if y_range[2] < 1:
        raise ValueError("The number of y samples must be >= 1.")

    if z_range[2] < 1:
        raise ValueError("The number of z samples must be >= 1.")

    f3d = autowrap_vectorfunction3d(function3d)
    x_samples = x_range[2]
    y_samples = y_range[2]
    z_samples = z_range[2]

    x = linspace(x_range[0], x_range[1], x_samples)
    y = linspace(y_range[0], y_range[1], y_samples)
    z = linspace(z_range[0], z_range[1], z_samples)
    v = empty((x_samples, y_samples, z_samples, 3))

    # obtain memoryviews for fast, direct memory access
    x_view = x
    y_view = y
    z_view = z
    v_view = v

    for i in range(x_samples):
        for j in range(y_samples):
            for k in range(z_samples):
                vector = f3d.evaluate(x_view[i], y_view[j], z_view[k])
                v_view[i, j, k, 0] = vector.x
                v_view[i, j, k, 1] = vector.y
                v_view[i, j, k, 2] = vector.z

    return x, y, z, v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray samplevector3d_points(object function3d, object points):
    """
    Sample a 3D vector function at the specified points.

    This function is for sampling at an unstructured sequence of points.
    For sampling over a regular grid, consider samplevector3d or
    samplevector3d_grid instead.

    :param function3d: a Python function or Function3D object
    :param points: an Nx3 array of points at which to sample the function
    :return: an Nx3 array containing the sampled values at each point

    .. code-block:: pycon

       >>> from raysect.core.math import Vector3D
       >>> from cherab.core.math import samplevector3d_points
       >>>
       >>> def my_func(x, y, z):
       >>>     return Vector3D(x, y, z)
       >>>
       >>> f_vals = samplevector3d_points(my_func, [[1,1,1], [2,2,2], [3,3,3]])
       >>> f_vals
       array([[1., 1., 1.],
              [2., 2., 2.],
              [3., 3., 3.]])
    """
    cdef:
        int i, j, nsamples
        VectorFunction3D f3d
        double[::1] x_view, y_view, z_view,
        double[:, ::1] v_view
        Vector3D vector

    points = asarray(points)
    if points.ndim != 2 or points.shape[1] != 3:
        raise ValueError("points should be an Nx3 array of points.")

    f3d = autowrap_vectorfunction3d(function3d)
    x = ascontiguousarray(points[:, 0], dtype=float)
    y = ascontiguousarray(points[:, 1], dtype=float)
    z = ascontiguousarray(points[:, 2], dtype=float)
    nsamples = points.shape[0]
    v = empty((nsamples, 3))

    x_view = x
    y_view = y
    z_view = z
    v_view = v

    for i in range(nsamples):
        vector = f3d.evaluate(x_view[i], y_view[i], z_view[i])
        v_view[i, 0] = vector.x
        v_view[i, 1] = vector.y
        v_view[i, 2] = vector.z

    return v


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray samplevector3d_grid(object function3d, object x, object y, object z):
    """
    Sample a 3D vector function on a rectilinear grid

    Note that v[i, j, k] = f(x[i], y[j], z[k])

    :param function3d: a Python function or Function3D object
    :param x: the x coordinates of each column in the grid
    :param y: the y coordinates of each row in the grid
    :param z: the z coordinates of each plane in the grid
    :return v: an NxMxkx3 array containing the sampled values at each grid point

    .. code-block:: pycon

       >>> from raysect.core.math import Vector3D
       >>> from cherab.core.math import samplevector3d_grid
       >>>
       >>> def my_func(x, y, z):
       >>>     return Vector3D(x, y, z)
       >>>
       >>> f_vals = samplevector3d_grid(my_func, [1,2], [1,2,3], [1,2,3])
       >>> f_vals
       array([[[[1., 1., 1.],
                [1., 1., 2.],
                [1., 1., 3.]],
               [[1., 2., 1.],
                [1., 2., 2.],
                [1., 2., 3.]],
               [[1., 3., 1.],
                [1., 3., 2.],
                [1., 3., 3.]]],

              [[[2., 1., 1.],
                [2., 1., 2.],
                [2., 1., 3.]],
               [[2., 2., 1.],
                [2., 2., 2.],
                [2., 2., 3.]],
               [[2., 3., 1.],
                [2., 3., 2.],
                [2., 3., 3.]]]])
    """
    cdef:
        int i, j, k, x_samples, y_samples, z_samples
        VectorFunction3D f3d
        double[::1] x_view, y_view, z_view
        double[:, :, :, ::1] v_view
        Vector3D vector

    x = ascontiguousarray(x, dtype=float)
    y = ascontiguousarray(y, dtype=float)
    z = ascontiguousarray(z, dtype=float)
    if x.ndim != 1:
        raise ValueError("x should be a 1D sequence of coordinates")
    if y.ndim != 1:
        raise ValueError("y should be a 1D sequence of coordinates")
    if z.ndim != 1:
        raise ValueError("z should be a 1D sequence of coordinates")

    f3d = autowrap_vectorfunction3d(function3d)

    x_samples = x.shape[0]
    y_samples = y.shape[0]
    z_samples = z.shape[0]
    v = empty((x_samples, y_samples, z_samples, 3))

    x_view = x
    y_view = y
    z_view = z
    v_view = v

    for i in range(x_samples):
        for j in range(y_samples):
            for k in range(z_samples):
                vector = f3d.evaluate(x_view[i], y_view[j], z_view[k])
                v_view[i, j, k, 0] = vector.x
                v_view[i, j, k, 1] = vector.y
                v_view[i, j, k, 2] = vector.z

    return v



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


cimport cython

from raysect.core.math.function.float cimport Function3D
from raysect.core.math.cython.utility cimport find_index

from libc.math cimport sqrt, exp, pi


cdef class ConstantAxisymmetricGaussian3D(Function3D):
    """
    A function with a 2D Gaussian in the x-y plane and equal standard deviations in x and y directions.

    .. math::
    F(x, y, z) = \\frac{1}{2 * \\pi \\sigma^2} exp\\left(-\\frac{x^2 + y^2}{2 * \\sigma^2}\\right)

    The function value has a Gaussian shape in the x-y plane with the standard deviations in
    x and y direction being equal. The integral over an x-y plane is equal to 1 
    and the mean values in x and y directions are equal to 0.

    :param float stddev: The standard deviation in both the x and y directions.
    """

    def __init__(self, stddev):

        super().__init__()

        self.stddev = stddev

    @property
    def stddev(self):

        return self._stddev

    @stddev.setter
    def stddev(self, value):
        if value <= 0:
            raise ValueError("Value has to be larger than 0.")

        self._stddev = value
        self._kr = -1 / (2 * value ** 2)
        self._normalisation = 1 / (2 * pi * value ** 2)

    cdef double evaluate(self, double x, double y, double z) except? -1e999:
        cdef:
            double r2
        r2 = x ** 2 + y ** 2
        return self._normalisation * exp(r2 * self._kr)


cdef class ConstantBivariateGaussian3D(Function3D):
    """
    A function with a 2D Gaussian in the x-y plane.

    .. math::
    F(x, y, z) = \\frac{1}{2 * \\pi \\sigma_x \\sigma_y} exp\\left(-\\frac{x^2 + y^2}{2 * \\sigma_x \\sigma_y}\\right)

    The function value has a Gaussian shape in the x-y plane. The integral over an x-y plane is equal to 1
    and the mean values in x and y directions are equal to 0.
    The correlation between the standard deviations in x and y directions is equal to 0.

    :param float stddev_x: The standard deviation in the x directions.
    :param float stddev_y: The standard deviation in the y directions.
    """

    def __init__(self, stddev_x, stddev_y):

        super().__init__()
        self._init_params()

        self.stddev_x = stddev_x
        self.stddev_y = stddev_y

    def _init_params(self):
        self._stddev_x = 1
        self._stddev_y = 1

    @property
    def stddev_x(self):
        return self._stddev_x

    @stddev_x.setter
    def stddev_x(self, value):
        if value <= 0:
            raise ValueError("Value has to be larger than 0.")

        self._stddev_x = value

        self._cache_constants()

    @property
    def stddev_y(self):
        return self._stddev_y

    @stddev_y.setter
    def stddev_y(self, value):
        if value <= 0:
            raise ValueError("Value has to be larger than 0.")

        self._stddev_y = value

        self._cache_constants()

    def _cache_constants(self):
        self._kx = -1 / (2 * self._stddev_x ** 2)
        self._ky = -1 / (2 * self._stddev_y ** 2)
        self._normalisation = 1 / (2 * pi * self._stddev_x * self._stddev_y)

    cdef double evaluate(self, double x, double y, double z) except? -1e999:
        return self._normalisation * exp(x ** 2 * self._kx +
                                                   y ** 2 * self._ky)


cdef class TrivariateGaussian3D(Function3D):
    """
    A function with a 3D Gaussian shape.

    .. math::
    F(x, y, z) = \\frac{1}{\\sqrt{2 \\pi^3} \\sigma_x \\sigma_y \\sigma_z} exp\\left(-\\frac{x^2}{2 \\sigma_x^2} -\\frac{y^2}{2 \\sigma_y^2} - \\frac{(z - \\mu_z)^2}{2 \\sigma_z^2}\\right)

    The integral over the whole 3D space is equal to 1.The correlation between the standard deviations in x and y directions is equal to 0. The mean value in the
    x and y directions are equal to 0.

    :param float mean_z: Mean value in the z direction.
    :param float stddev_x: The standard deviation in the x directions.
    :param float stddev_y: The standard deviation in the y directions.
    :param float stddev_z: The standard deviation in the z directions.
    """

    def __init__(self, mean_z, stddev_x, stddev_y, stddev_z):

        super().__init__()
        self._init_params()

        self.stddev_x = stddev_x
        self.stddev_y = stddev_y
        self.stddev_z = stddev_z
        self.mean_z = mean_z

    def _init_params(self):
        self._mean_z = 1
        self._stddev_x = 1
        self._stddev_y = 1
        self._stddev_z = 1

    @property
    def stddev_x(self):
        return self._stddev_x

    @stddev_x.setter
    def stddev_x(self, double value):
        if value <= 0:
            raise ValueError("Value has to be larger than 0.")

        self._stddev_x = value

        self._cache_constants()

    @property
    def stddev_y(self):
        return self._stddev_y

    @stddev_y.setter
    def stddev_y(self, double value):
        if value <= 0:
            raise ValueError("Value has to be larger than 0.")

        self._stddev_y = value

        self._cache_constants()

    @property
    def stddev_z(self):
        return self._stddev_z

    @stddev_z.setter
    def stddev_z(self, double value):
        if value <= 0:
            raise ValueError("Value has to be larger than 0.")

        self._stddev_z = value

        self._cache_constants()

    @property
    def mean_z(self):
        return self._mean_z

    @mean_z.setter
    def mean_z(self, double value):
        self._mean_z = value

    def _cache_constants(self):
        self._kx = -1 / (2 * self._stddev_x ** 2)
        self._ky = -1 / (2 * self._stddev_y ** 2)
        self._kz = -1 / (2 * self._stddev_z ** 2)
        self._normalisation = 1 / (sqrt((2 * pi) ** 3) * self._stddev_x * self._stddev_y * self._stddev_z)

    cdef double evaluate(self, double x, double y, double z) except? -1e999:
        return self._normalisation * exp(x ** 2 * self._kx +
                                         y ** 2 * self._ky +
                                        (z - self._mean_z) ** 2 * self._kz)


cdef class GaussianBeamModel(Function3D):
    """
    A Gaussian beam function (https://en.wikipedia.org/wiki/Gaussian_beam)

    .. math::
      F(x, y, z) = \\frac{1}{2 \\pi \\sigma^2_z} exp\\left( -\\frac{x^2 + y^2}{2 \\sigma_z(z)^2 }\\right)

    where the standard deviation in the z direction

    .. math::
      \\sigma_z(z) = \\sigma_0 \\sqrt{1 + \\left(\\frac{z - z_0}{z_R}\\right)^2}

    is a function of position and the

    .. math::
      z_R = \\frac{\\pi \\omega_0^2 n}{\\lambda_l}

    is the Rayleigh range.
    """

    def __init__(self, double wavelength, double waist_z, double stddev_waist):

        # preset default values
        self._wavelength = 1e3
        self._waist_z = 0
        self._stddev_waist = 1e-3

        self.wavelength = wavelength
        self.waist_z = waist_z
        self.stddev_waist = stddev_waist

    @property
    def wavelength(self):
        return self._wavelength

    @wavelength.setter
    def wavelength(self, double value):
        if value <= 0:
            raise ValueError("Value has to be larger than 0, but {0} passed.".format(value))

        self._wavelength = value
        self._cache_constants()

    @property
    def waist_z(self):
        return self._waist_z

    @waist_z.setter
    def waist_z(self, double value):
        self._waist_z = value

    @property
    def stddev_waist(self):
        return self._stddev_waist

    @stddev_waist.setter
    def stddev_waist(self, double value):
        if value <= 0:
            raise ValueError("Value has to be larger than 0, but {0} passed.".format(value))

        self._stddev_waist = value
        self._stddev_waist2 = self._stddev_waist ** 2
        self._cache_constants()

    def _cache_constants(self):

        n = 1  # refractive index of vacuum
        self._rayleigh_range = 2 * pi * n * self._stddev_waist2 / self._wavelength / 1e-9

    @cython.cdivision(True)
    cdef double evaluate(self, double x, double y, double z) except? -1e999:

        cdef:
            double r2, stddev_z2, z_prime

        # shift to correct gaussiam beam model coords, it works with waist at z=0
        z_prime = z - self._waist_z

        r2 = x ** 2 + y ** 2
        stddev_z2 = self._stddev_waist2 * (1 + ((z_prime) / self._rayleigh_range) ** 2)

        return 1 / (2 * pi * stddev_z2) * exp(r2 / (-2 * stddev_z2))



# Copyright 2016-2022 Euratom
# Copyright 2016-2022 United Kingdom Atomic Energy Authority
# Copyright 2016-2022 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
cimport numpy as np
from libc.math cimport cos, sin
from raysect.primitive.mesh.mesh cimport Mesh
from raysect.core.math.polygon import triangulate2d
from raysect.core.math.cython cimport winding2d


cdef double DEG2RAD = 2 * np.pi / 360


cpdef Mesh toroidal_mesh_from_polygon(object polygon, double toroidal_extent, object polygon_triangles=None, int num_toroidal_segments=500):
    """
    Generates a watertight Raysect Mesh primitive from the specified 2D polygon in R-Z plane
    by extending it in toroidal direction by a given angle and closing the
    poloidal faces with triangulated polygons.

    :param object polygon: An object which can be converted to a numpy array with shape 
                           [N,2] specifying the wall outline polygon in the R-Z plane. 
                           The polygon should not be closed, i.e. vertex i = 0 and i = N 
                           should not be the same vertex, but neighbours.
    :param object polygon_triangles: An object which can be converted to a numpy array
                                     with shape [M,3] specifying the triangulation
                                     of a polygon (polygon_triangles = [[v1, v2, v3],...),
                                     where v1, v2, v3 are the vertex array indices specifying
                                     the triangle’s vertices. Should be with clockwise winding.
                                     Defaults to None.
                                     If not provided, the triangulation will be performed using
                                     `triangulate2d(polygon)` from raysect.core.math.polygon.
    :param float toroidal_extent: Angular extention of an element in toroidal direction (in degrees).
                                  Note that in the case of toroidal_extent=360  produces 
                                  an axisymmetric mesh which has no end faces
    :param int num_toroidal_segments: The number of repeating toroidal segments
                                      per given `toroidal_extent` that will be used to construct
                                      the mesh. Defaults to 500.

    :return: A watertight Raysect Mesh primitive constructed from the R-Z polygon.

    .. code-block:: pycon

        >>> from cherab.tools.primitives import toroidal_mesh_from_polygon
        >>>
        >>> # wall_polygon is your (N, 2) ndarray describing the polygon
        >>> mesh = toroidal_mesh_from_polygon(wall_polygon, toroidal_extent = 50)
    """

    cdef:
        int num_poloidal_vertices, num_toroidal_segments_loop
        int i, j, vid, v1_id, v2_id, v3_id, v4_id
        double theta, r, x, y, z
        list triangles
        double[:, :] polygon_mv, vertices_mv

    polygon = np.array(polygon, dtype=np.float64)

    if polygon.ndim != 2:
        raise ValueError("The 'polygon' must be a two-dimensional array.")

    if polygon.shape[0] < 2:
        raise ValueError("The 'polygon' must contain at least two vertices.")

    if polygon.shape[1] != 2:
        raise ValueError("The 'polygon' must have [N, 2] shape.")
    
    if (toroidal_extent<=0) or (toroidal_extent>360):
        raise ValueError("The 'toroidal_extent' should be in the range 0 < 'toroidal_extent' <= 360.")

    num_poloidal_vertices = polygon.shape[0]
    theta = toroidal_extent / num_toroidal_segments  # toroidal step

    vertices = np.zeros((num_poloidal_vertices * num_toroidal_segments, 3))
    vertices_mv = vertices
    polygon_mv = polygon

    for i in range(num_toroidal_segments):
        for j in range(num_poloidal_vertices):

            r = polygon_mv[j, 0]
            z = polygon_mv[j, 1]
            x = r * cos(i * theta * DEG2RAD)
            y = r * sin(i * theta * DEG2RAD)

            vid = i * num_poloidal_vertices + j
            vertices_mv[vid, 0] = x
            vertices_mv[vid, 1] = y
            vertices_mv[vid, 2] = z

    # assemble mesh triangles
    triangles = []

    if toroidal_extent != 360.:
        if polygon_triangles is None:
            # triangulating initial polygon in case of not full axisymmetric mesh
            polygon_triangles = triangulate2d(polygon)

        else:
            # user-defined triangulation
            polygon_triangles = np.array(polygon_triangles, dtype=np.int32)
            # check data sanity
            _check_polygon_triangulation(polygon, polygon_triangles)

        triangles = triangles + polygon_triangles[:, [1, 0, 2]].tolist()

    num_toroidal_segments_loop = num_toroidal_segments if toroidal_extent == 360. else num_toroidal_segments - 1

    for i in range(num_toroidal_segments_loop):
        for j in range(num_poloidal_vertices):

            if i == num_toroidal_segments - 1 and j == num_poloidal_vertices - 1:
                v1_id = i * num_poloidal_vertices + j
                v2_id = i * num_poloidal_vertices + 0
                v3_id = j
                v4_id = 0

            elif i == num_toroidal_segments - 1:
                v1_id = i * num_poloidal_vertices + j
                v2_id = i * num_poloidal_vertices + j + 1
                v3_id = j
                v4_id = j + 1

            elif j == num_poloidal_vertices - 1:
                v1_id = i * num_poloidal_vertices + j
                v2_id = i * num_poloidal_vertices
                v3_id = i * num_poloidal_vertices + num_poloidal_vertices + j
                v4_id = i * num_poloidal_vertices + num_poloidal_vertices

            else:
                v1_id = i * num_poloidal_vertices + j
                v2_id = i * num_poloidal_vertices + j + 1
                v3_id = i * num_poloidal_vertices + num_poloidal_vertices + j
                v4_id = i * num_poloidal_vertices + num_poloidal_vertices + j + 1

            triangles.append([v1_id, v2_id, v4_id])
            triangles.append([v4_id, v3_id, v1_id])

    if toroidal_extent != 360.:
        triangles = triangles + ((num_toroidal_segments - 1) * num_poloidal_vertices + polygon_triangles).tolist()

    return Mesh(vertices=vertices, triangles=triangles, smoothing=False)


cdef _check_polygon_triangulation(np.ndarray polygon, np.ndarray polygon_triangles):


    # check consistency
    if not np.array_equal(np.unique(polygon_triangles), np.arange(polygon.shape[0], dtype=np.int32)):
        raise ValueError("The data in 'polygon_triangles' does not match a given 'polygon'.")
    # check winding
    triangles = np.ascontiguousarray(polygon[polygon_triangles])
    if not all(winding2d(triangle) for triangle in triangles):
        raise ValueError("All triangles in 'polygon_triangles' must be wound clockwise.")



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


from raysect.core.math.function.float cimport Function3D
from raysect.core.math.function.vector3d cimport Function3D as VectorFunction3D

from raysect.optical cimport Spectrum, Point3D, Vector3D

from cherab.core.laser.node cimport Laser


cdef class LaserProfile:

    cdef:
        VectorFunction3D _polarization3d, _pointing3d
        Function3D _energy_density3d
        readonly object notifier

    cpdef Vector3D get_pointing(self, double x, double y, double z)

    cpdef Vector3D get_polarization(self, double x, double y, double z)

    cpdef double get_energy_density(self, double x, double y, double z)

    cpdef list generate_geometry(self)


# -*- coding: utf-8 -*-
#
# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
#
# The following code is created by Vladislav Neverov (NRC "Kurchatov Institute") for Cherab Spectroscopy Modelling Framework

"""
Very simple but fast pipelines for ray transfer matrix (geometry matrix) calculation.
When calculating the ray transfer matrix, the spectral array is used to store the radiance
from individual unit light sources and not the actual spectrum. In this case the spectral
array may contain ~ 10,000 spectral bins but the wavelengths for all of them are equal.
Spectral pipelines from Raysect still can be used, but they are slower compared to ray
transfer pipelines. Standard error is not calculated in these pipelines, only the mean value.
Dispersive rendering and adaptive sampling features are removed to improve the performance.
"""

import numpy as np
from raysect.optical.observer.base import Pipeline0D, Pipeline1D, Pipeline2D, PixelProcessor


class RayTransferPipelineBase():

    def __init__(self, name=None, kind='power'):

        self.name = name
        self._matrix = None
        self._samples = 0
        self._bins = 0
        self.kind = kind

    @property
    def kind(self):
        """
        The kind of the pipeline. Can be 'power' or 'radiance'.
        In the case of 'power', the resulting matrix is multiplied by the sensitivity
        of the detector, and the units of the matrix are [m^3 sr], which gives the units
        of power [W] for the product of the ray transfer matrix and the emission profile.
        In case of 'radiance', the sensitivity is not taken into account and
        the matrix is calculated in [m], which gives the units of radiance [W m^-2 sr^-1]
        for the product of the ray transfer matrix and the emission profile.
        """
        return self._kind

    @kind.setter
    def kind(self, value):
        _kind = value.lower()
        if _kind in ('power', 'radiance'):
            self._kind = _kind
        else:
            raise ValueError("The kind property must be 'power' or 'radiance'.")

    @property
    def matrix(self):
        return self._matrix


class RayTransferPipeline0D(Pipeline0D, RayTransferPipelineBase):
    """
    Simple 0D pipeline for ray transfer matrix (geometry matrix) calculation.

    :param str name: The name of the pipeline. Default is 'RayTransferPipeline0D'.
    :param str kind: The kind of the pipeline. Can be 'power' (default) or 'radiance'.
        In the case of 'power', the resulting matrix is multiplied by the sensitivity
        of the detector, and the units of the matrix are [m^3 sr], which gives the units
        of power [W] for the product of the ray transfer matrix and the emission profile.
        In case of 'radiance', the sensitivity is not taken into account and
        the matrix is calculated in [m], which gives the units of radiance [W m^-2 sr^-1]
        for the product of the ray transfer matrix and the emission profile.
        Note that if the sensitivity of the detector is 1 (e.g. `PinholeCamera`, `VectorCamera`),
        the 'power' and 'radiance' give the same results.

    :ivar np.ndarray matrix: Ray transfer matrix, a 1D array of size :math:`N_{bin}`.

    .. code-block:: pycon

       >>> from cherab.tools.raytransfer import RayTransferPipeline0D
       >>> pipeline = RayTransferPipeline0D(kind='radiance')
    """

    def __init__(self, name='RayTransferPipeline0D', kind='power'):

        RayTransferPipelineBase.__init__(self, name, kind)

    def initialise(self, min_wavelength, max_wavelength, spectral_bins, spectral_slices, quiet):
        self._samples = 0
        self._bins = spectral_bins
        self._matrix = np.zeros(spectral_bins)

    def pixel_processor(self, slice_id):
        if self._kind == 'power':
            return PowerRayTransferPixelProcessor(self._bins)
        else:
            return RadianceRayTransferPixelProcessor(self._bins)

    def update(self, slice_id, packed_result, pixel_samples):
        self._samples += pixel_samples
        self._matrix += packed_result[0]

    def finalise(self):
        self._matrix /= self._samples


class RayTransferPipeline1D(Pipeline1D, RayTransferPipelineBase):
    """
    Simple 1D pipeline for ray transfer matrix (geometry matrix) calculation.

    :param str name: The name of the pipeline. Default is 'RayTransferPipeline0D'.
    :param str kind: The kind of the pipeline. Can be 'power' (default) or 'radiance'.
        In the case of 'power', the resulting matrix is multiplied by the sensitivity
        of the detector, and the units of the matrix are [m^3 sr], which gives the units
        of power [W] for the product of the ray transfer matrix and the emission profile.
        In case of 'radiance', the sensitivity is not taken into account and
        the matrix is calculated in [m], which gives the units of radiance [W m^-2 sr^-1]
        for the product of the ray transfer matrix and the emission profile.
        Note that if the sensitivity of the detector is 1 (e.g. `PinholeCamera`, `VectorCamera`),
        the 'power' and 'radiance' give the same results.

    :ivar np.ndarray matrix: Ray transfer matrix, a 2D array of shape :math:`(N_{pixel}, N_{bin})`.

    .. code-block:: pycon

       >>> from cherab.tools.raytransfer import RayTransferPipeline1D
       >>> pipeline = RayTransferPipeline1D(kind='radiance')
    """

    def __init__(self, name='RayTransferPipeline1D', kind='power'):

        RayTransferPipelineBase.__init__(self, name, kind)
        self._pixels = None

    def initialise(self, pixels, pixel_samples, min_wavelength, max_wavelength, spectral_bins, spectral_slices, quiet):
        self._pixels = pixels
        self._samples = pixel_samples
        self._bins = spectral_bins
        self._matrix = np.zeros((pixels, spectral_bins))

    def pixel_processor(self, pixel, slice_id):
        if self._kind == 'power':
            return PowerRayTransferPixelProcessor(self._bins)
        else:
            return RadianceRayTransferPixelProcessor(self._bins)

    def update(self, pixel, slice_id, packed_result):
        self._matrix[pixel] = packed_result[0] / self._samples

    def finalise(self):
        pass


class RayTransferPipeline2D(Pipeline2D, RayTransferPipelineBase):
    """
    Simple 2D pipeline for ray transfer matrix (geometry matrix) calculation.

    :param str name: The name of the pipeline. Default is 'RayTransferPipeline0D'.
    :param str kind: The kind of the pipeline. Can be 'power' (default) or 'radiance'.
        In the case of 'power', the resulting matrix is multiplied by the sensitivity
        of the detector, and the units of the matrix are [m^3 sr], which gives the units
        of power [W] for the product of the ray transfer matrix and the emission profile.
        In case of 'radiance', the sensitivity is not taken into account and
        the matrix is calculated in [m], which gives the units of radiance [W m^-2 sr^-1]
        for the product of the ray transfer matrix and the emission profile.
        Note that if the sensitivity of the detector is 1 (e.g. `PinholeCamera`, `VectorCamera`),
        the 'power' and 'radiance' give the same results.

    :ivar np.ndarray matrix: Ray transfer matrix, a 3D array of shape :math:`(N_x, N_y, N_{bin})`.

    .. code-block:: pycon

       >>> from cherab.tools.raytransfer import RayTransferPipeline2D
       >>> pipeline = RayTransferPipeline2D(kind='radiance')
    """

    def __init__(self, name='RayTransferPipeline2D', kind='power'):

        RayTransferPipelineBase.__init__(self, name, kind)
        self._pixels = None

    def initialise(self, pixels, pixel_samples, min_wavelength, max_wavelength, spectral_bins, spectral_slices, quiet):
        self._pixels = pixels
        self._samples = pixel_samples
        self._bins = spectral_bins
        self._matrix = np.zeros((pixels[0], pixels[1], spectral_bins))

    def pixel_processor(self, x, y, slice_id):
        if self._kind == 'power':
            return PowerRayTransferPixelProcessor(self._bins)
        else:
            return RadianceRayTransferPixelProcessor(self._bins)

    def update(self, x, y, slice_id, packed_result):
        self._matrix[x, y] = packed_result[0] / self._samples

    def finalise(self):
        pass


class RayTransferPixelProcessorBase(PixelProcessor):
    """
    Base class for PixelProcessor that stores ray transfer matrix for each pixel.
    """

    def __init__(self, bins):
        self._matrix = np.zeros(bins)

    def pack_results(self):
        return (self._matrix, 0)


class RadianceRayTransferPixelProcessor(RayTransferPixelProcessorBase):
    """
    PixelProcessor that stores ray transfer matrix in the units of [m] for each pixel.
    """

    def add_sample(self, spectrum, sensitivity):
        self._matrix += spectrum.samples


class PowerRayTransferPixelProcessor(RayTransferPixelProcessorBase):
    """
    PixelProcessor that stores ray transfer matrix in the units of [m^3 sr] for each pixel.
    """

    def add_sample(self, spectrum, sensitivity):
        self._matrix += spectrum.samples * sensitivity



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.optical cimport Node, Primitive, AffineMatrix3D
from raysect.optical.material.emitter.inhomogeneous cimport VolumeIntegrator

from cherab.core.atomic cimport AtomicData, Element
from cherab.core.distribution cimport DistributionFunction
from cherab.core.species cimport Species
from cherab.core.math cimport VectorFunction3D
from cherab.core.plasma.model cimport PlasmaModel


cdef class Composition:

    cdef:
        dict _species
        readonly object notifier

    cpdef object set(self, object species)

    cpdef object add(self, Species species)

    cpdef Species get(self, Element element, int charge)

    cpdef object clear(self)


cdef class ModelManager:

    cdef:
        list _models
        readonly object notifier

    cpdef object set(self, object models)

    cpdef object add(self, PlasmaModel model)

    cpdef object clear(self)


cdef class Plasma(Node):

    cdef:

        readonly object notifier
        VectorFunction3D _b_field
        DistributionFunction _electron_distribution
        Composition _composition
        AtomicData _atomic_data
        Primitive _geometry
        AffineMatrix3D _geometry_transform
        ModelManager _models
        VolumeIntegrator _integrator

    cdef object __weakref__

    cdef VectorFunction3D get_b_field(self)

    cdef DistributionFunction get_electron_distribution(self)

    cdef Composition get_composition(self)

    cpdef double z_effective(self, double x, double y, double z) except -1

    cpdef double ion_density(self, double x, double y, double z)






# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import unittest

import numpy as np

from cherab.core.utility import conversion as cv


class FactorConversion(cv.BaseFactorConversion):
    conversion_factor = 42.


class TestConversion(unittest.TestCase):

    def setUp(self):
        self.array = np.array([1., 0., 2.5e6, 4.7e-3, 6.3e19, 9.2e-15])

    def tearDown(self):
        del self.array

    def test_evamutoms_to_array(self):
        converted_array = cv.EvAmuToMS.to(self.array.copy())
        for i in range(len(self.array)):
            self.assertAlmostEqual(converted_array[i], cv.EvAmuToMS.to(self.array[i]), delta=1e-10,
                                   msg='EvAmuToMS forward conversion failed to convert values given in a numpy array (failure from index {})!'.format(i))

    def test_evamutoms_inv_array(self):
        converted_array = cv.EvAmuToMS.inv(self.array.copy())
        for i in range(len(self.array)):
            self.assertAlmostEqual(converted_array[i], cv.EvAmuToMS.inv(self.array[i]), delta=1e-10,
                                   msg='EvAmuToMS backward conversion failed to convert values given in a numpy array (failure from index {})!'.format(i))

    def test_photontoj_to_array(self):
        wvl = 529.
        converted_array = cv.PhotonToJ.to(self.array.copy(), wvl)
        for i in range(len(self.array)):
            self.assertAlmostEqual(converted_array[i], cv.PhotonToJ.to(self.array[i], wvl), delta=1e-10,
                                   msg='PhotonToJ forward conversion failed to convert values given in a numpy array (failure from index {})!'.format(i))

    def test_photontoj_inv_array(self):
        wvl = 529.
        converted_array = cv.PhotonToJ.inv(self.array.copy(), wvl)
        for i in range(len(self.array)):
            self.assertAlmostEqual(converted_array[i], cv.PhotonToJ.inv(self.array[i], wvl), delta=1e-10,
                                   msg='PhotonToJ backward conversion failed to convert values given in a numpy array (failure from index {})!'.format(i))

    def test_factorconversion_to_array(self):
        converted_array = FactorConversion.to(self.array.copy())
        for i in range(len(self.array)):
            self.assertAlmostEqual(converted_array[i], FactorConversion.to(self.array[i]), delta=1e-10,
                                   msg='Forward conversion derived from BaseFactorConversion failed to convert values given in a numpy array (failure from index {})!'.format(i))

    def test_factorconversion_inv_array(self):
        converted_array = FactorConversion.inv(self.array.copy())
        for i in range(len(self.array)):
            self.assertAlmostEqual(converted_array[i], FactorConversion.inv(self.array[i]), delta=1e-10,
                                   msg='Backward conversion derived from BaseFactorConversion failed to convert values given in a numpy array (failure from index {})!'.format(i))



if __name__ == '__main__':
    unittest.main()



# Copyright 2016-2024 Euratom
# Copyright 2016-2024 United Kingdom Atomic Energy Authority
# Copyright 2016-2024 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


import os
import json
import numpy as np

from cherab.core.atomic import Element
from cherab.core.utility import RecursiveDict
from .utility import DEFAULT_REPOSITORY_PATH, valid_charge


def add_line_power_rate(species, charge, rate, repository_path=None):
    """
    Adds a single line radiated power rate to the repository.

    If adding multiple rates, consider using the update_line_power_rates()
    function instead. The update function avoids repeatedly opening and closing
    the rate files.

    :param species: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param rate: Line radiated power rate dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'rate': array-like of size (N, M) with line radiated power rate in W.m^3.

    :param repository_path: Path to the atomic data repository.
    """

    update_line_power_rates({
        species: {
            charge: rate
        }
    }, repository_path)


def update_line_power_rates(rates, repository_path=None):
    """
    Update the files for the line radiated power rates:
    /radiated_power/line/<species>.json
    in the atomic data repository.

    File contains multiple rates, indexed by the ion's charge state.

    :param rates: Dictionary in the form {<species>: {<charge>: <rate>}}, where

    |      <species> is the plasma species (Element/Isotope),
    |      <charge> is the charge of the plasma species,
    |      <rate> is the line radiated rate dictionary containing the following entries:
    |          'ne': array-like of size (N) with electron density in m^-3,
    |          'te': array-like of size (M) with electron temperature in eV,
    |          'rate': array-like of size (N, M) with line radiated power rate in W.m^3.

    :param repository_path: Path to the atomic data repository.
    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for species, rate_data in rates.items():

        # sanitise and validate arguments
        if not isinstance(species, Element):
            raise TypeError('The species must be an Element object.')

        path = os.path.join(repository_path, 'radiated_power/line/{}.json'.format(species.symbol.lower()))

        _update_and_write_adf11(species, rate_data, path)


def add_continuum_power_rate(species, charge, rate, repository_path=None):
    """
    Adds a single continuum power rate to the repository.

    If adding multiple rates, consider using the update_continuum_power_rates()
    function instead. The update function avoids repeatedly opening and closing
    the rate files.

    :param species: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param rate: Continuum power rate dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'rate': array-like of size (N, M) with continuum power rate in W.m^3.

    :param repository_path: Path to the atomic data repository.
    """

    update_line_power_rates({
        species: {
            charge: rate
        }
    }, repository_path)


def update_continuum_power_rates(rates, repository_path=None):
    """
    Update the files for the continuum power rates:
    /radiated_power/continuum/<species>.json
    in the atomic data repository.

    File contains multiple rates, indexed by ion's charge state.

    :param rates: Dictionary in the form {<species>: {<charge>: <rate>}}, where

    |      <species> is the plasma species (Element/Isotope),
    |      <charge> is the charge of the plasma species,
    |      <rate> is the continuum power rate dictionary containing the following entries:
    |          'ne': array-like of size (N) with electron density in m^-3,
    |          'te': array-like of size (M) with electron temperature in eV,
    |          'rate': array-like of size (N, M) with continuum power rate in W.m^3.

    :param repository_path: Path to the atomic data repository.
    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for species, rate_data in rates.items():

        # sanitise and validate arguments
        if not isinstance(species, Element):
            raise TypeError('The species must be an Element object.')

        path = os.path.join(repository_path, 'radiated_power/continuum/{}.json'.format(species.symbol.lower()))

        _update_and_write_adf11(species, rate_data, path)


def add_cx_power_rate(species, charge, rate, repository_path=None):
    """
    Adds a single CX radiation power rate to the repository
    (charge exchage with neutral hydrogen).

    If adding multiple rates, consider using the update_cx_power_rates()
    function instead. The update function avoids repeatedly opening and closing
    the rate files.

    :param species: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param rate: CX power rate dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'rate': array-like of size (N, M) with CX power rate in W.m^3.

    :param repository_path: Path to the atomic data repository.
    """

    update_line_power_rates({
        species: {
            charge: rate
        }
    }, repository_path)


def update_cx_power_rates(rates, repository_path=None):
    """
    Update the files for the CX radiation power rates
    (charge exchage with neutral hydrogen):
    /radiated_power/cx/<species>.json
    in the atomic data repository.

    File contains multiple rates, indexed by ion's charge state.

    :param rates: Dictionary in the form {<species>: {<charge>: <rate>}}, where

    |      <species> is the plasma species (Element/Isotope),
    |      <charge> is the charge of the plasma species,
    |      <rate> is the thermal CX power rate dictionary containing the following entries:
    |          'ne': array-like of size (N) with electron density in m^-3,
    |          'te': array-like of size (M) with electron temperature in eV,
    |          'rate': array-like of size (N, M) with thermal CX power rate in W.m^3.

    :param repository_path: Path to the atomic data repository.    
    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for species, rate_data in rates.items():

        # sanitise and validate arguments
        if not isinstance(species, Element):
            raise TypeError('The species must be an Element object.')

        path = os.path.join(repository_path, 'radiated_power/cx/{}.json'.format(species.symbol.lower()))

        _update_and_write_adf11(species, rate_data, path)


def _update_and_write_adf11(species, rate_data, path):

    # read in any existing rates
    try:
        with open(path, 'r') as f:
            content = RecursiveDict.from_dict(json.load(f))
    except FileNotFoundError:
        content = RecursiveDict()

    for charge, rates in rate_data.items():

        if not valid_charge(species, charge):
            raise ValueError('The charge state is larger than the number of protons in the specified species.')

        # sanitise and validate rate data
        te = np.array(rates['te'], np.float64)
        ne = np.array(rates['ne'], np.float64)
        rate_table = np.array(rates['rates'], np.float64)

        if ne.ndim != 1:
            raise ValueError('Density array must be a 1D array.')

        if te.ndim != 1:
            raise ValueError('Temperature array must be a 1D array.')

        if (ne.shape[0], te.shape[0]) != rate_table.shape:
            raise ValueError('Electron temperature, density and rate data arrays have inconsistent sizes.')

        # update file content with new rate
        content[str(charge)] = {
            'te': te.tolist(),
            'ne': ne.tolist(),
            'rate': rate_table.tolist(),
        }

        # create directory structure if missing
        directory = os.path.dirname(path)
        if not os.path.isdir(directory):
            os.makedirs(directory)

        # write new data
        with open(path, 'w') as f:
            json.dump(content, f, indent=2, sort_keys=True)


def get_line_radiated_power_rate(element, charge, repository_path=None):
    """
    Reads the line radiated power rate for the given species and charge
    from the atomic data repository.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param repository_path: Path to the atomic data repository.

    :return rate: Line radiated power rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with line radiated power rate in W.m^3.

    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    path = os.path.join(repository_path, 'radiated_power/line/{}.json'.format(element.symbol.lower()))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[str(charge)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested radiated power rate (element={}, charge={})'
                           ' is not available.'.format(element.symbol, charge))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d


def get_continuum_radiated_power_rate(element, charge, repository_path=None):
    """
    Reads the continuum power rate for the given species and charge
    from the atomic data repository.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param repository_path: Path to the atomic data repository.

    :return rate: Continuum power rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with continuum power rate in W.m^3.

    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    path = os.path.join(repository_path, 'radiated_power/continuum/{}.json'.format(element.symbol.lower()))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[str(charge)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested radiated power rate (element={}, charge={})'
                           ' is not available.'.format(element.symbol, charge))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d


def get_cx_radiated_power_rate(element, charge, repository_path=None):
    """
    Reads the CX radiation power rate for the given species and charge
    from the atomic data repository.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param repository_path: Path to the atomic data repository.

    :return rate: CX radiation power rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with CX radiation power rate in W.m^3.

    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    path = os.path.join(repository_path, 'radiated_power/cx/{}.json'.format(element.symbol.lower()))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[str(charge)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested radiated power rate (element={}, charge={})'
                           ' is not available.'.format(element.symbol, charge))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
cimport numpy as np

from cherab.core.math.function cimport autowrap_function1d, Function1D

cimport cython

np.import_array()

DEF MULTIPLET_WAVELENGTH = 0
DEF MULTIPLET_RATIO = 1

DEF PI_POLARISATION = 0
DEF SIGMA_PLUS_POLARISATION = 1
DEF SIGMA_MINUS_POLARISATION = -1

cdef class ZeemanStructure():
    r"""
    Provides wavelengths and ratios of
    :math:`\pi`-/:math:`\sigma`-polarised Zeeman components for any given value of
    magnetic field strength.

    :param list pi_components: A list of 2-tuples of Function1D objects that provide the
                               wavelengths and ratios of individual :math:`\pi`-polarised
                               Zeeman components for a given magnetic field strength:
                               [(wvl_func1, ratio_func1), (wvl_func2, ratio_func2), ...]
    :param list sigma_plus_components: A list of 2-tuples of Function1D objects that provide the
                                       wavelengths and ratios of individual
                                       :math:`\sigma^{+}`-polarised Zeeman components for
                                       a given magnetic field strength:
                                       [(wvl_func1, ratio_func1), (wvl_func2, ratio_func2), ...]
    :param list sigma_minus_components: A list of 2-tuples of Function1D objects that provide the
                                        wavelengths and ratios of individual
                                        :math:`\sigma^{-}`-polarised
                                        Zeeman components for a given magnetic field strength:
                                        [(wvl_func1, ratio_func1), (wvl_func2, ratio_func2), ...]
    """

    def __init__(self, pi_components, sigma_plus_components, sigma_minus_components):

        cdef tuple component

        self._pi_components = []
        self._sigma_plus_components = []
        self._sigma_minus_components = []

        for component in pi_components:
            if len(component) != 2:
                raise ValueError('Argument "pi_components" must be a list of 2-tuples.')
            self._pi_components.append((autowrap_function1d(component[MULTIPLET_WAVELENGTH]),
                                        autowrap_function1d(component[MULTIPLET_RATIO])))

        for component in sigma_plus_components:
            if len(component) != 2:
                raise ValueError('Argument "sigma_plus_components" must be a list of 2-tuples.')
            self._sigma_plus_components.append((autowrap_function1d(component[MULTIPLET_WAVELENGTH]),
                                                autowrap_function1d(component[MULTIPLET_RATIO])))

        for component in sigma_minus_components:
            if len(component) != 2:
                raise ValueError('Argument "sigma_minus_components" must be a list of 2-tuples.')
            self._sigma_minus_components.append((autowrap_function1d(component[MULTIPLET_WAVELENGTH]),
                                                 autowrap_function1d(component[MULTIPLET_RATIO])))

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    @cython.cdivision(True)
    cdef double[:, :] evaluate(self, double b, int polarisation):

        cdef int i
        cdef np.npy_intp multiplet_shape[2]
        cdef double ratio_sum
        cdef np.ndarray multiplet
        cdef double[:, :] multiplet_mv
        cdef Function1D wavelength, ratio
        cdef list components

        if b < 0:  # safety check in case if used stand-alone
            raise ValueError('Argument "b" (magnetic field strength) must be non-negative.')

        if polarisation == PI_POLARISATION:
            components = self._pi_components
        elif polarisation == SIGMA_PLUS_POLARISATION:
            components = self._sigma_plus_components
        elif polarisation == SIGMA_MINUS_POLARISATION:
            components = self._sigma_minus_components
        else:
            raise ValueError('Argument "polarisation" must be 0, 1 or -1, {} given.'.format(polarisation))

        multiplet_shape[0] = 2
        multiplet_shape[1] = len(components)
        multiplet = np.PyArray_SimpleNew(2, multiplet_shape, np.NPY_FLOAT64)
        multiplet_mv = multiplet

        ratio_sum = 0
        for i in range(len(components)):
            wavelength = components[i][MULTIPLET_WAVELENGTH]
            ratio = components[i][MULTIPLET_RATIO]

            multiplet_mv[MULTIPLET_WAVELENGTH, i] = wavelength.evaluate(b)
            multiplet_mv[MULTIPLET_RATIO, i] = ratio.evaluate(b)

            ratio_sum += multiplet_mv[MULTIPLET_RATIO, i]

        # normalising ratios
        if ratio_sum > 0:
            for i in range(multiplet_mv.shape[1]):
                multiplet_mv[MULTIPLET_RATIO, i] /= ratio_sum

        return multiplet_mv

    def __call__(self, double b, str polarisation):

        if polarisation.lower() == 'pi':
            return np.asarray(self.evaluate(b, PI_POLARISATION))

        if polarisation.lower() == 'sigma_plus':
            return np.asarray(self.evaluate(b, SIGMA_PLUS_POLARISATION))

        if polarisation.lower() == 'sigma_minus':
            return np.asarray(self.evaluate(b, SIGMA_MINUS_POLARISATION))

        raise ValueError('Argument "polarisation" must be "pi", "sigma_plus" or "sigma_minus", {} given.'.fotmat(polarisation))



# Copyright 2016-2024 Euratom
# Copyright 2016-2024 United Kingdom Atomic Energy Authority
# Copyright 2016-2024 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


import os
import json
import numpy as np

from cherab.core.atomic import Element
from cherab.core.utility import RecursiveDict
from .utility import DEFAULT_REPOSITORY_PATH, valid_charge


def add_ionisation_rate(species, charge, rate, repository_path=None):
    """
    Adds a single ionisation rate to the repository.

    If adding multiple rates, consider using the update_ionisation_rates()
    function instead. The update function avoids repeatedly opening and closing
    the rate files.

    :param species: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param rate: Ionisation rate dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'rate': array-like of size (N, M) with ionisation rate in m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    update_ionisation_rates({
        species: {
            charge: rate
        }
    }, repository_path)


def update_ionisation_rates(rates, repository_path=None):
    """
    Updates the ionisation rate files `/ionisation/<species>.json`
    in atomic data repository.

    File contains multiple rates, indexed by the ion charge state.

    :param rates: Dictionary in the form {<species>: {<charge>: <rate>}}, where

    |      <species> is the plasma species (Element/Isotope),
    |      <charge> is the charge of the plasma species,
    |      <rate> is the ionisation rate dictionary containing the following entries:
    |          'ne': array-like of size (N) with electron density in m^-3,
    |          'te': array-like of size (M) with electron temperature in eV,
    |          'rate': array-like of size (N, M) with ionisation rate in m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for species, rate_data in rates.items():

        # sanitise and validate arguments
        if not isinstance(species, Element):
            raise TypeError('The species must be an Element object.')

        path = os.path.join(repository_path, 'ionisation/{}.json'.format(species.symbol.lower()))

        _update_and_write_adf11(species, rate_data, path)


def add_recombination_rate(species, charge, rate, repository_path=None):
    """
    Adds a single recombination rate to the repository.

    If adding multiple rates, consider using the update_recombination_rates()
    function instead. The update function avoids repeatedly opening and closing
    the rate files.

    :param species: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param rate: Recombination rate dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'rate': array-like of size (N, M) with recombination rate in m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    update_recombination_rates({
        species: {
            charge: rate
        }
    }, repository_path)


def update_recombination_rates(rates, repository_path=None):
    """
    Updates the recombination rate files `/recombination/<species>.json`
    in the atomic data repository.

    File contains multiple rates, indexed by the ion charge state.

    :param rates: Dictionary in the form {<species>: {<charge>: <rate>}}, where

    |      <species> is the plasma species (Element/Isotope),
    |      <charge> is the charge of the plasma species,
    |      <rate> is the recombination rate dictionary containing the following entries:
    |          'ne': array-like of size (N) with electron density in m^-3,
    |          'te': array-like of size (M) with electron temperature in eV,
    |          'rate': array-like of size (N, M) with recombination rate in m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for species, rate_data in rates.items():

        # sanitise and validate arguments
        if not isinstance(species, Element):
            raise TypeError('The species must be an Element object.')

        path = os.path.join(repository_path, 'recombination/{}.json'.format(species.symbol.lower()))

        _update_and_write_adf11(species, rate_data, path)


def add_thermal_cx_rate(donor_element, donor_charge, receiver_element, rate, repository_path=None):
    """
    Adds a single thermal charge exchange rate to the repository.

    If adding multiple rates, consider using the update_recombination_rates()
    function instead. The update function avoids repeatedly opening and closing
    the rate files.

    :param donor_element: Element donating the electron.
    :param donor_charge: Charge of the donating atom/ion.
    :param receiver_element: Element receiving the electron.
    :param receiver_charge: Charge of the receiving atom/ion.
    :param rate: Thermal CX rate dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'rate': array-like of size (N, M) with thermal CX rate in m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    rates2update = RecursiveDict()
    rates2update[donor_element][donor_charge][receiver_element] = rate

    update_thermal_cx_rates(rates2update, repository_path)


def update_thermal_cx_rates(rates, repository_path=None):
    """
    Updates the thermal charge exchange rate files
    `/thermal_cx/<donor_element>/<donor_charge>/<receiver_element>.json`
    in the atomic data repository.

    File contains multiple rates, indexed by the ion charge state.

    :param rates: Dictionary in the form:

    |  { <donor_element>: { <donor_charge>: { <receiver_element>: { <donor_charge>: <rate> } } } }, where
    |      <donor_element> is the element donating the electron.
    |      <donor_charge> is the charge of the donating atom/ion.
    |      <receiver_element> is the element receiving the electron.
    |      <receiver_charge> is the charge of the receiving atom/ion.
    |      <rate> is the thermal CX rate dictionary containing the following entries:
    |          'ne': array-like of size (N) with electron density in m^-3,
    |          'te': array-like of size (M) with electron temperature in eV,
    |          'rate': array-like of size (N, M) with thermal CX rate in m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for donor_element in rates.keys():
        for donor_charge in rates[donor_element].keys():
            for receiver_element, rate_data in rates[donor_element][donor_charge].items():

                # sanitise and validate arguments
                if not isinstance(receiver_element, Element):
                    raise TypeError('The receiver_element must be an Element object.')

                rate_path = 'thermal_cx/{0}/{1}/{2}.json'.format(donor_element.symbol.lower(),
                                                                 donor_charge, receiver_element.symbol.lower())
                path = os.path.join(repository_path, rate_path)

                _update_and_write_adf11(receiver_element, rate_data, path)


def _update_and_write_adf11(species, rate_data, path):

        # read in any existing rates
        try:
            with open(path, 'r') as f:
                content = RecursiveDict.from_dict(json.load(f))
        except FileNotFoundError:
            content = RecursiveDict()

        for charge, rates in rate_data.items():

            if not valid_charge(species, charge):
                raise ValueError('Charge state is larger than the number of protons in the specified species.')

            # sanitise and validate rate data
            te = np.array(rates['te'], np.float64)
            ne = np.array(rates['ne'], np.float64)
            rate_table = np.array(rates['rates'], np.float64)

            if ne.ndim != 1:
                raise ValueError('Density array must be a 1D array.')

            if te.ndim != 1:
                raise ValueError('Temperature array must be a 1D array.')

            if (ne.shape[0], te.shape[0]) != rate_table.shape:
                raise ValueError('Electron temperature, density and rate data arrays have inconsistent sizes.')

            # update file content with new rate
            content[str(charge)] = {
                'te': te.tolist(),
                'ne': ne.tolist(),
                'rate': rate_table.tolist(),
            }

            # create directory structure if missing
            directory = os.path.dirname(path)
            if not os.path.isdir(directory):
                os.makedirs(directory)

            # write new data
            with open(path, 'w') as f:
                json.dump(content, f, indent=2, sort_keys=True)


def get_ionisation_rate(element, charge, repository_path=None):
    """
    Reads the ionisation rate for the given species and charge
    from the atomic data repository.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param repository_path: Path to the atomic data repository.

    :return rate: Ionisation rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with ionisation rate in m^3.s^-1.

    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    path = os.path.join(repository_path, 'ionisation/{}.json'.format(element.symbol.lower()))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[str(charge)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested ionisation rate (element={}, charge={})'
                           ' is not available.'.format(element.symbol, charge))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d


def get_recombination_rate(element, charge, repository_path=None):
    """
    Reads the recombination rate for the given species and charge
    from the atomic data repository.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param repository_path: Path to the atomic data repository.

    :return rate: Recombination rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with recombination rate in m^3.s^-1.

    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    path = os.path.join(repository_path, 'recombination/{}.json'.format(element.symbol.lower()))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[str(charge)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested recombination rate (element={}, charge={})'
                           ' is not available.'.format(element.symbol, charge))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d


def get_thermal_cx_rate(donor_element, donor_charge, receiver_element, receiver_charge, repository_path=None):
    """
    Reads the thermal charge exchange rate for the given species and charge
    from the atomic data repository.

    :param donor_element: Element donating the electron.
    :param donor_charge: Charge of the donating atom/ion.
    :param receiver_element: Element receiving the electron.
    :param receiver_charge: Charge of the receiving atom/ion.
    :param repository_path: Path to the atomic data repository.

    :return rate: Thermal CX rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with thermal CX rate in m^3.s^-1.

    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    rate_path = 'thermal_cx/{0}/{1}/{2}.json'.format(donor_element.symbol.lower(), donor_charge,
                                                     receiver_element.symbol.lower())
    path = os.path.join(repository_path, rate_path)
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[str(receiver_charge)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested thermal charge-exchange rate (donor={}, donor charge={}, receiver={})'
                           ' is not available.'
                           ''.format(donor_element.symbol, donor_charge, receiver_element.symbol, receiver_charge))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from numpy import ndarray
from raysect.optical.observer import FibreOptic

from .base import Observer0DGroup


class FibreOpticGroup(Observer0DGroup):
    """
    A group of fibre optics under a single scene-graph node.

    A scene-graph object regrouping a series of 'FibreOptic'
    observers as a scene-graph parent. Allows combined observation and display
    control simultaneously.

    :ivar list/float acceptance_angle: The angle in degrees between the z axis and the cone
                                       surface which defines the fibres solid angle sampling
                                       area. The same value can be shared between all observers,
                                       or each observer can be assigned with individual value.
    :ivar list/float radius: The radius of the fibre tip in metres. This radius defines a circular
                             area at the fibre tip which will be sampled over. The same value
                             can be shared between all observers, or each observer can be
                             assigned with individual value.

    .. code-block:: pycon

       >>> from math import cos, sin, pi
       >>>
       >>> import matplotlib.pyplot as plt
       >>> from raysect.core import translate, rotate_basis, Point3D, Vector3D
       >>> from raysect.optical import World
       >>> from raysect.optical.observer import RadiancePipeline0D, SpectralRadiancePipeline0D, PowerPipeline0D, SpectralPowerPipeline0D, FibreOptic
       >>>
       >>> from cherab.tools.observers import FibreOpticGroup
       >>> from cherab.tools.observers.group.plotting import plot_group_total, plot_group_spectra
       >>>
       >>> world = World()
       >>>
       >>> transform1 = translate(3., 0, 0) * rotate_basis(Vector3D(-cos(pi/10), 0, sin(pi/10)), Vector3D(0, 1, 0))
       >>> fibre1 = FibreOptic(transform=transform1, name="Fibre 1")
       >>> transform2 = translate(3, 0 ,0) * rotate_basis(Vector3D(-1, 0, 0), Vector3D(0, 1, 0))
       >>> fibre2 = FibreOptic(transform=transform2, name="Fibre 2")
       >>> transform3 = translate(3, 0, 0) * rotate_basis(Vector3D(-cos(pi/10), 0, -sin(pi/10)), Vector3D(0, 1, 0))
       >>> fibre3 = FibreOptic(transform=transform3, name="Fibre 3")
       >>>
       >>> group = FibreOpticGroup(name='MyFibreGroup', parent=world, observers=[fibre1, fibre2])
       >>> group.add_observer(fibre3)
       >>> pipelines = [SpectralRadiancePipeline0D, RadiancePipeline0D]
       >>> keywords = [{'name': 'MySpectralPipeline'}, {'name': 'MyMonoPipeline'}]
       >>> group.connect_pipelines(pipelines, keywords)  # add pipelines to all observers in the group
       >>> group.acceptance_angle = 2  # same value for all fibres in the group
       >>> group.radius = 2.e-3
       >>> group.spectral_bins = 512
       >>> group.pixel_samples = [2000, 1000, 2000]  # individual value for each fibre in the group
       >>> group.observe()  # combined observation
       >>>
       >>> plot_group_spectra(group, item='MySpectralPipeline', in_photons=True)  # plot the spectra
       >>> plot_group_total(group, item='MyMonoPipeline')  # plot the total signals
       >>> plt.show()
    """
    _OBSERVER_TYPE = FibreOptic

    @property
    def acceptance_angle(self):
        # The angle in degrees between the z axis and the cone surface which defines the fibres
        # solid angle sampling area.
        return [observer.acceptance_angle for observer in self._observers]

    @acceptance_angle.setter
    def acceptance_angle(self, value):
        if isinstance(value, (list, tuple, ndarray)):
            if len(value) == len(self._observers):
                for observer, v in zip(self._observers, value):
                    observer.acceptance_angle = v
            else:
                raise ValueError("The length of 'acceptance_angle' ({}) "
                                 "mismatches the number of observers ({}).".format(len(value), len(self._observers)))
        else:
            for observer in self._observers:
                observer.acceptance_angle = value

    @property
    def radius(self):
        # The radius of the fibre tip in metres. This radius defines a circular area at the fibre tip
        # which will be sampled over.
        return [observer.radius for observer in self._observers]

    @radius.setter
    def radius(self, value):
        if isinstance(value, (list, tuple, ndarray)):
            if len(value) == len(self._observers):
                for observer, v in zip(self._observers, value):
                    observer.radius = v
            else:
                raise ValueError("The length of 'radius' ({}) "
                                 "mismatches the number of observers ({}).".format(len(value), len(self._observers)))
        else:
            for observer in self._observers:
                observer.radius = value



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from numpy cimport ndarray
from cherab.core.math.function cimport Function1D


cdef class _Interpolate1DBase(Function1D):

    cdef:
        double[::1] _x, _f
        bint _constant
        int _extrapolation_type
        double _extrapolation_range

    cdef object _build(self, ndarray x, ndarray f)

    cdef double evaluate(self, double px) except? -1e999

    cpdef double derivative(self, double px, int order) except? -1e999

    cdef double _evaluate(self, double px, int order, int index) except? -1e999

    cdef double _extrapolate(self, double px, int order, int index, double rx) except? -1e999

    cdef double _extrapol_linear(self, double px, int order, int index, double rx) except? -1e999

    cdef double _extrapol_quadratic(self, double px, int order, int index, double rx) except? -1e999


cdef class Interpolate1DLinear(_Interpolate1DBase):
    pass


cdef class Interpolate1DCubic(_Interpolate1DBase):

    cdef:
        int _continuity_order
        double _ox, _sx, _of, _sf
        double[:,::1] _k

    cdef double _calc_polynomial_derivative(self, int i, double p, int order)



# Copyright 2016-2022 Euratom
# Copyright 2016-2022 United Kingdom Atomic Energy Authority
# Copyright 2016-2022 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from os import path
import numpy as np
import json

from libc.math cimport log10, log, M_PI, sqrt
from raysect.core.math.function.float cimport Interpolator2DArray
from cherab.core.utility.constants cimport RYDBERG_CONSTANT_EV, SPEED_OF_LIGHT, ELEMENTARY_CHARGE, PLANCK_CONSTANT

cimport cython


DEF EULER_GAMMA = 0.5772156649015329

cdef double PH_TO_EV_FACTOR = PLANCK_CONSTANT * SPEED_OF_LIGHT * 1e9 / ELEMENTARY_CHARGE


cdef class FreeFreeGauntFactor():
    """
    The base class for temperature-averaged free-free Gaunt factors.
    """

    cpdef double evaluate(self, double z, double temperature, double wavelength) except? -1e999:
        """
        Returns the temperature-averaged free-free Gaunt factor for the supplied parameters.

        :param double z: Species charge or effective plasma charge.
        :param double temperature: Electron temperature in eV.
        :param double wavelength: Spectral wavelength.

        :return: free-free Gaunt factor
        """
        raise NotImplementedError("The evaluate() virtual method must be implemented.")

    def __call__(self, double z, double temperature, double wavelength):
        """
        Returns the temperature-averaged free-free Gaunt factor for the supplied parameters.

        :param double z: Species charge or effective plasma charge.
        :param double temperature: Electron temperature in eV.
        :param double wavelength: Spectral wavelength.

        :return: free-free Gaunt factor
        """

        return self.evaluate(z, temperature, wavelength)


cdef class InterpolatedFreeFreeGauntFactor(FreeFreeGauntFactor):
    r"""
    The temperature-averaged free-free Gaunt factors interpolated in the space of parameters:
    :math:`u = h{\nu}/kT` and :math:`{\gamma}^{2} = Z^{2}Ry/kT`.
    See T.R. Carson, 1988, Astron. & Astrophys., 189,
    `319 <https://ui.adsabs.harvard.edu/#abs/1988A&A...189..319C/abstract>`_ for details.

    The cubic interpolation in a semi-log space is used.

    The Born approximation is used outside the interpolation range.

    :param object u: A 1D array-like object of real values.
    :param object gamma2: A 1D array-like object of real values.
    :param object gaunt_factor: 2D array-like object of real values
                                storing the Gaunt factor values at u, gamma2.

    :ivar tuple u_range: The interpolation range of `u` parameter.
    :ivar tuple gamma2_range: The interpolation range of :math:`\\gamma^2` parameter.
    :ivar dict raw_data: Dictionary containing the raw data.
    """

    def __init__(self, object u, object gamma2, object gaunt_factor):

        u = np.array(u, dtype=np.float64)
        u.flags.writeable = False
        gamma2 = np.array(gamma2, dtype=np.float64)
        gamma2.flags.writeable = False
        gaunt_factor = np.array(gaunt_factor, dtype=np.float64)
        gaunt_factor.flags.writeable = False

        self.raw_data = {'u': u, 'gamma2': gamma2, 'gaunt_factor': gaunt_factor}

        self._u_min = u.min()
        self._u_max = u.max()
        self._gamma2_min = gamma2.min()
        self._gamma2_max = gamma2.max()

        self.u_range = (self._u_min, self._u_max)
        self.gamma2_range = (self._gamma2_min, self._gamma2_max)

        self._gaunt_factor = Interpolator2DArray(np.log10(u), np.log10(gamma2), gaunt_factor, 'cubic', 'none', 0, 0)

    @cython.cdivision(True)
    cpdef double evaluate(self, double z, double temperature, double wavelength) except? -1e999:
        """
        Returns the temperature-averaged free-free Gaunt factor for the supplied parameters.

        :param double z: Species charge or effective plasma charge.
        :param double temperature: Electron temperature in eV.
        :param double wavelength: Spectral wavelength.

        :return: free-free Gaunt factor
        """

        cdef:
            double u, gamma2

        if z == 0:

            return 0

        gamma2 = z * z * RYDBERG_CONSTANT_EV / temperature
        u = PH_TO_EV_FACTOR / (temperature * wavelength)

        # classical limit
        if u >= self._u_max or gamma2 >= self._gamma2_max:

            return 1

        # Born approximation limit
        if u < self._u_min or gamma2 < self._gamma2_min:

            return sqrt(3) / M_PI * (log(4 / u) - EULER_GAMMA)

        return self._gaunt_factor.evaluate(log10(u), log10(gamma2))


cdef class MaxwellianFreeFreeGauntFactor(InterpolatedFreeFreeGauntFactor):
    r"""
    The Maxwellian-averaged free-free Gaunt factor interpolated over the data from Table A.1 in
    M.A. de Avillez and D. Breitschwerdt, "Temperature-averaged and total free-free Gaunt factors
    for κ and Maxwellian distributions of electrons", 2015, Astron. & Astrophys. 580,
    `A124 <https://www.aanda.org/articles/aa/full_html/2015/08/aa26104-15/aa26104-15.html>`_.

    The Born approximation is used outside the interpolation range.
    """

    def __init__(self):

        with open(path.join(path.dirname(__file__), "data/maxwellian_free_free_gaunt_factor.json")) as f:
            data = json.load(f)

        super().__init__(data['u'], data['gamma2'], data['gaunt_factor'])



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


from raysect.core.math.function.float cimport Function3D

cdef class ConstantAxisymmetricGaussian3D(Function3D):

    cdef:
        double _stddev, _normalisation, _kr

    cdef double evaluate(self, double x, double y, double z) except? -1e999


cdef class ConstantBivariateGaussian3D(Function3D):

    cdef:
        double _stddev_x, _stddev_y, _kx, _ky, _normalisation


cdef class TrivariateGaussian3D(Function3D):

    cdef:
        double _mean_z, _stddev_x, _stddev_y, _stddev_z, _kx, _ky
        double _kz, _normalisation


cdef class GaussianBeamModel(Function3D):

    cdef:
        double _waist_z, _stddev_waist, _stddev_waist2, _wavelength, _rayleigh_range

    cdef double evaluate(self, double x, double y, double z) except? -1e999



import unittest

import numpy as np

from cherab.core.math.caching import Caching1D


class TestCaching1D(unittest.TestCase):

    def setUp(self):
        self.function = lambda x: np.cos(10 * x)
        self.grad_f = lambda x: -10 * np.sin(10 * x)
        self.hess_f = lambda x: 100 * np.cos(10 * x)
        xmin, xmax = -10, 2
        self.space_area = xmin, xmax
        self.resolution = 0.1

    def tearDown(self):
        del self.function
        del self.grad_f
        del self.space_area
        del self.resolution

    def tolerance(self, x):
        tolerance = self.grad_f(x) * self.resolution + self.hess_f(x) * self.resolution ** 2 * 0.5
        return max(abs(tolerance), 0.1)

    def test_values(self):
        cached_func = Caching1D(self.function, self.space_area, self.resolution)

        for x in np.linspace(self.space_area[0], self.space_area[1], 100):
            self.assertAlmostEqual(cached_func(x), self.function(x), delta=self.tolerance(x),
                                   msg='Cached function at {} is too far from exact function!'.format(x))
            self.assertAlmostEqual(cached_func(x), self.function(x), delta=0.1,
                                   msg='Cached function at {} is too far from exact function!'.format(x))


if __name__ == '__main__':
    unittest.main()



# Copyright 2016-2024 Euratom
# Copyright 2016-2024 United Kingdom Atomic Energy Authority
# Copyright 2016-2024 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
from libc.math cimport INFINITY, log10

from raysect.core.math.function.float cimport Interpolator2DArray


cdef class LineRadiationPower(CoreLineRadiationPower):
    """
    Line radiated power coefficient.

    The data is interpolated with cubic spline in log-log space.
    Nearest neighbour extrapolation is used if extrapolate is True.

    :param Element species: Element object defining the ion type.
    :param int ionisation: Charge state of the ion.
    :param dict data: Line radiated power rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with radiated power rate in W.m^3.

    :param bint extrapolate: Enable extrapolation (default=False).

    :ivar tuple density_range: Electron density interpolation range.
    :ivar tuple temperature_range: Electron temperature interpolation range.
    :ivar dict raw_data: Dictionary containing the raw data.
    """

    def __init__(self, species, ionisation, dict data, extrapolate=False):

        super().__init__(species, ionisation)

        self.raw_data = data

        # unpack
        ne = data['ne']
        te = data['te']
        rate = np.log10(data['rate'])

        # store limits of data
        self.density_range = ne.min(), ne.max()
        self.temperature_range = te.min(), te.max()

        # interpolate rate
        # using nearest extrapolation to avoid infinite values at 0 for some rates
        extrapolation_type = 'nearest' if extrapolate else 'none'
        self._rate = Interpolator2DArray(np.log10(ne), np.log10(te), rate, 'cubic', extrapolation_type, INFINITY, INFINITY)

    cpdef double evaluate(self, double electron_density, double electron_temperature) except? -1e999:

        # need to handle zeros, also density and temperature can become negative due to cubic interpolation
        if electron_density <= 0 or electron_temperature <= 0:
            return 0

        # calculate rate and convert from log10 space to linear space
        return 10 ** self._rate.evaluate(log10(electron_density), log10(electron_temperature))


cdef class NullLineRadiationPower(CoreLineRadiationPower):
    """
    A line radiation power rate that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double electron_density, double electron_temperature) except? -1e999:
        return 0.0


cdef class ContinuumPower(CoreContinuumPower):
    """
    Recombination continuum radiated power coefficient.

    The data is interpolated with cubic spline in log-log space.
    Nearest neighbour extrapolation is used if extrapolate is True.

    :param Element species: Element object defining the ion type.
    :param int ionisation: Charge state of the ion.
    :param dict data: Recombination continuum radiated power rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with radiated power rate in W.m^3.

    :param bint extrapolate: Enable extrapolation (default=False).

    :ivar tuple density_range: Electron density interpolation range.
    :ivar tuple temperature_range: Electron temperature interpolation range.
    :ivar dict raw_data: Dictionary containing the raw data.
    """

    def __init__(self, species, ionisation, dict data, extrapolate=False):

        super().__init__(species, ionisation)

        self.raw_data = data

        # unpack
        ne = data['ne']
        te = data['te']
        rate = np.log10(data['rate'])

        # store limits of data
        self.density_range = ne.min(), ne.max()
        self.temperature_range = te.min(), te.max()

        # interpolate rate
        # using nearest extrapolation to avoid infinite values at 0 for some rates
        extrapolation_type = 'nearest' if extrapolate else 'none'
        self._rate = Interpolator2DArray(np.log10(ne), np.log10(te), rate, 'cubic', extrapolation_type, INFINITY, INFINITY)

    cpdef double evaluate(self, double electron_density, double electron_temperature) except? -1e999:

        # need to handle zeros, also density and temperature can become negative due to cubic interpolation
        if electron_density <= 0 or electron_temperature <= 0:
            return 0

        # calculate rate and convert from log10 space to linear space
        return 10 ** self._rate.evaluate(log10(electron_density), log10(electron_temperature))


cdef class NullContinuumPower(CoreContinuumPower):
    """
    A continuum radiation power rate that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double electron_density, double electron_temperature) except? -1e999:
        return 0.0


cdef class CXRadiationPower(CoreCXRadiationPower):
    """
    Charge exchange radiated power coefficient.

    The data is interpolated with cubic spline in log-log space.
    Linear extrapolation is used if extrapolate is True.

    :param Element species: Element object defining the ion type.
    :param int ionisation: Charge state of the ion.
    :param dict data: CX radiated power rate dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with radiated power rate in W.m^3.

    :param bint extrapolate: Enable extrapolation (default=False).

    :ivar tuple density_range: Electron density interpolation range.
    :ivar tuple temperature_range: Electron temperature interpolation range.
    :ivar dict raw_data: Dictionary containing the raw data.
    """

    def __init__(self, species, ionisation, dict data, extrapolate=False):

        super().__init__(species, ionisation)

        self.raw_data = data

        # unpack
        ne = data['ne']
        te = data['te']
        rate = np.log10(data['rate'])

        # store limits of data
        self.density_range = ne.min(), ne.max()
        self.temperature_range = te.min(), te.max()

        # interpolate rate
        extrapolation_type = 'linear' if extrapolate else 'none'
        self._rate = Interpolator2DArray(np.log10(ne), np.log10(te), rate, 'cubic', extrapolation_type, INFINITY, INFINITY)

    cpdef double evaluate(self, double electron_density, double electron_temperature) except? -1e999:

        # need to handle zeros, also density and temperature can become negative due to cubic interpolation
        if electron_density <= 0 or electron_temperature <= 0:
            return 0

        # calculate rate and convert from log10 space to linear space
        return 10 ** self._rate.evaluate(log10(electron_density), log10(electron_temperature))


cdef class NullCXRadiationPower(CoreCXRadiationPower):
    """
    A CX radiation power rate that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double electron_density, double electron_temperature) except? -1e999:
        return 0.0



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from numpy import ndarray
from raysect.optical.observer import SightLine

from .base import Observer0DGroup


class SightLineGroup(Observer0DGroup):
    """
    A group of SightLine under a single scene-graph node.

    A scene-graph object regrouping a series of 'SightLine'
    observers as a scene-graph parent. Allows combined observation and display
    control simultaneously.

    :ivar list observers: A list of sight lines (SightLine instances) in this group.

    .. code-block:: pycon

       >>> from math import cos, sin, pi
       >>>
       >>> import matplotlib.pyplot as plt
       >>> from raysect.core import translate, rotate_basis, Point3D, Vector3D
       >>> from raysect.optical import World
       >>> from raysect.optical.observer import RadiancePipeline0D, SpectralRadiancePipeline0D, PowerPipeline0D, SpectralPowerPipeline0D, SightLine
       >>>
       >>> from cherab.tools.observers import SightLineGroup
       >>> from cherab.tools.observers.group.plotting import plot_group_total, plot_group_spectra
       >>>
       >>> world = World()
       >>>
       >>> transform1 = translate(3., 0, 0) * rotate_basis(Vector3D(-cos(pi/10), 0, sin(pi/10)), Vector3D(0, 1, 0))
       >>> sightline_1 = SightLine(transform=transform1, name="SightLine 1")
       >>> transform2 = translate(3, 0 ,0) * rotate_basis(Vector3D(-1, 0, 0), Vector3D(0, 1, 0))
       >>> sightline_2 = SightLine(transform=transform2, name="SightLine 2")
       >>> transform3 = translate(3, 0, 0) * rotate_basis(Vector3D(-cos(pi/10), 0, -sin(pi/10)), Vector3D(0, 1, 0))
       >>> sightline_3 = SightLine(transform=transform3, name="SightLine 3")
       >>>
       >>> group = SightLineGroup(name='MySightLineGroup', parent=world, observers=[sightline_1, sightline_2])
       >>> group.add_observer(sightline_3)
       >>> pipelines = [SpectralRadiancePipeline0D, RadiancePipeline0D]
       >>> keywords = [{'name': 'MySpectralPipeline'}, {'name': 'MyMonoPipeline'}]
       >>> group.connect_pipelines(pipelines, keywords)  # add pipelines to all observers in the group
       >>> group.acceptance_angle = 2  # same value for all sightlines in the group
       >>> group.radius = 2.e-3
       >>> group.spectral_bins = 512
       >>> group.pixel_samples = [2000, 1000, 2000]  # individual value for each sightline in the group
       >>> group.observe()  # combined observation
       >>>
       >>> plot_group_spectra(group, item='MySpectralPipeline', in_photons=True)  # plot the spectra
       >>> plot_group_total(group, item='MyMonoPipeline')  # plot the total signals
       >>> plt.show()
    """
    _OBSERVER_TYPE = SightLine

    @property
    def sensitivity(self):
        # A list of observer sensitivities.
        return [observer.sensitivity for observer in self._observers]

    @sensitivity.setter
    def sensitivity(self, value):
        if isinstance(value, (list, tuple, ndarray)):
            if len(value) == len(self._observers):
                for observer, v in zip(self._observers, value):
                    observer.sensitivity = v
            else:
                raise ValueError("The length of 'sensitivity' ({}) ".format(len(value)) +
                                 "mismatches the number of observers ({}).".format(len(self._observers)))
        else:
            for observer in self._observers:
                observer.sensitivity = value
